#include <hex.h>
#include <curses.h>
#include <time.h>
#include <ctype.h>
#include <stdarg.h>
#include <signal.h>
#include <setjmp.h>

WINDOW *screen = NULL, *top, *data, *status;

extern byte *fname;
extern off_t text_size;
extern byte *text;

extern off_t cursor;
extern char cursor_type;
extern int modified;

static sigjmp_buf func_env;

static int help_page = 0;
static off_t sel_start = 1, sel_end = 0;
static off_t page_start = 0;

static void type_char(int c);
static void show_line(off_t start, int line);
static void show_data(off_t start);
static void page_up(void);
static void page_down(void);
static void parse_control(void);
static void sigint_break(int x);

int init_screen(void)
{
	screen = initscr();
	if (!screen) return -1;
	
	top = newwin(1, COLS, 0, 0);
	data = newwin(LINES - 4, COLS, 2, 0);
	status = newwin(1, COLS, LINES - 1, 0);

	if (!top || !data || !status) {
		endwin();
		return -1;
	}

	scrollok(top, FALSE); scrollok(data, TRUE); scrollok(status, FALSE);
	cbreak(); noecho(); nodelay(data, TRUE);
	meta(data, TRUE); keypad(data, TRUE);

	return 0;
}

void show_top(int modified)
{
	time_t t;
	struct tm *tb;
	unsigned char tmp[1024];

	memset(tmp, '-', 511);
	tmp[512] = 0;

	t = time(NULL);
	tb = localtime(&t);
	sprintf(tmp + COLS - 11, "[ %.2d:%02d ]--",
		tb->tm_hour, tb->tm_min);
	wmove(top, 0, 0); waddstr(top, tmp);

	sprintf(tmp, "---[ P H E X ]---[ %s%s ]---[ %lu KB ]",
		fname, modified ? " (modified)" : "", text_size / 1024);
	tmp[512] = 0;
	
	wmove(top, 0, 0); waddstr(top, tmp);

	wmove(top, 0, 5); waddch(top, 'P' | A_BOLD);
	wmove(top, 0, 7); waddch(top, 'H' | A_BOLD);
	wmove(top, 0, 9); waddch(top, 'E' | A_BOLD);
	wmove(top, 0, 11); waddch(top, 'X' | A_BOLD);
	
	wrefresh(top);
}

void show_line(off_t start, int line)
{
	off_t curr_off;
	int j, k, c;
	char tmp[128];
	
	wmove(data, line, 2);
	wprintw(data, "%07lX:  ", start + line * 16);

	tmp[0] = 0;

	for (j = 0; j < 16; j++) {
		curr_off = start + line * 16 + j;
		if (curr_off < text_size) {
			sprintf(tmp, "%02X ", text[curr_off]);
		} else {
			sprintf(tmp, "   ");
		}
		if (j == 7) strcat(tmp, " ");
		for (k = 0; tmp[k]; k++)
			if ((curr_off >= sel_start) && (curr_off <= sel_end))
				waddch(data, tmp[k] | A_REVERSE);
			else
				waddch(data, tmp[k]);
	}
	waddch(data, ' ');
	for (j = 0; j < 16; j++) {
		curr_off = start + line * 16 + j;
		if (curr_off < text_size) {
			c = (int) text[curr_off];
		} else {
			c = ' ';
		}
		if (!isprint(c)) {
			c = '.';
		}
		if ((curr_off >= sel_start) && (curr_off <= sel_end))
			waddch(data, c | A_REVERSE);
		else
			waddch(data, c);
	}
}

static void show_data(off_t start)
{
	int i;
	
	for (i = 0; i < (LINES - 4); i++) {
		show_line(start, i);
	}
	page_start = start;
	wrefresh(data);
}

void move_cursor(off_t new)
{
	int curs_x = 0; int curs_y = 0;
	int page = (LINES - 4) * 16;

	if (new >= text_size) new = text_size - 1;

	/* speed up long searches */
	if (new > (page_start + page * 3))
		page_start = (new - new % 16) - page * 3;
	if (new < (page_start - page * 3))
		page_start = (new - new % 16) + page * 3;
	
	if ((new >= page_start) && ((new - page_start) < page)) {
		curs_y = (new - page_start) / 16;
		curs_x = (new - page_start) % 16;
		if (cursor_type == CURSOR_HEX) {
			curs_x = curs_x * 3 + 12;
			if (curs_x > 32) curs_x++;
			wmove(data, curs_y, curs_x);
		} else {
			curs_x += 62;
			wmove(data, curs_y, curs_x);
		}
		wrefresh(data);
		cursor = new;
		return;
	}
	if ((new - page_start) >= page) {
		wscrl(data, 1);
		page_start += 16;
		show_line(page_start, LINES - 5);
		move_cursor(new);
	}
	if (new < page_start) {
		wscrl(data, -1);
		page_start -= 16;
		show_line(page_start, 0);
		move_cursor(new);
	}
}

static void page_up(void)
{
	if (cursor < ((LINES - 4) * 16)) {
		move_cursor(0);
	} else {
		move_cursor(cursor - (LINES - 4) * 16);
	}
}

static void page_down(void)
{
	if ((text_size - cursor) <= ((LINES - 4) * 16)) {
		move_cursor(text_size - 1);
	} else {
		move_cursor(cursor + (LINES - 4) * 16);
	}
}

static void parse_control(void)
{
	int c;
	
	wmove(top, 0, 0);
	waddch(top, '^' | A_BOLD | A_BLINK);
	waddch(top, 'K' | A_BOLD | A_BLINK);
	wrefresh(top);
	move_cursor(cursor);

	while (1) {
		usleep(10000);
		c = wgetch(data);
		if (c == ERR) continue;

		wmove(top, 0, 0);
		waddstr(top, "--");
		wrefresh(top);
		move_cursor(cursor);

  		switch(c) {
			case 'h':
				print_help(0);
				break;
			case 'u':
				move_cursor(0); break;
			case 'v':
				move_cursor(text_size - 1); break;
			case 'b':
				sel_start = cursor; show_data(page_start);
				move_cursor(cursor);
				break;
			case 'k':
				sel_end = cursor; show_data(page_start);
				move_cursor(cursor);
				break;
			case 'l':
				goto_offset(); break;
			case 'f':
				snr_txt(); break;
			case 'n':
				move_cursor(bfi_search()); break;
			case 'd':
				save_file(); break;
			case 'w':
				save_block(sel_start, sel_end); break;
			case 'e':
				edit_file(); show_data(page_start);
				move_cursor(cursor); break;
			case 'r':
				insert_file(); show_data(page_start);
				move_cursor(cursor); break;
			case 'c':
				copy_block(sel_start, sel_end);
				show_data(page_start); move_cursor(cursor);
				break;
		}
		break;
	}
}

void main_loop(void)
{
	int c;
	
	c = wgetch(data);
	
	if (c == ERR) return;
	
	switch (c) {
		case 9: /* tab switches hex <-> char mode */
			cursor_type = (cursor_type + 1) % 2;
			move_cursor(cursor);
			return;

		case KEY_LEFT: /* go left */
			if (cursor) move_cursor(cursor - 1);
			return;

		case KEY_RIGHT: /* go right */
			if (cursor < (text_size - 1))
				move_cursor(cursor + 1);
			return;

		case KEY_UP: /* go up */
			if (cursor < 16)
				move_cursor(0);
			else
				move_cursor(cursor - 16);
			return;

		case KEY_DOWN: /* go down */
			if (cursor < (text_size - 16)) 
				move_cursor(cursor + 16);
			else
				move_cursor(text_size - 1);
			return;

		case KEY_HOME:
			move_cursor(cursor - cursor % 16);
			return;

		case KEY_END:
			if ((cursor - cursor % 16) < (text_size - 15))
				move_cursor(cursor - cursor % 16 + 15);
			else
				move_cursor(text_size - 1);
			return;

		case KEY_PPAGE:
			page_up();
			return;

		case KEY_NPAGE:
			page_down();
			return;

		case 21:
			page_up();
			return;
		
		case 22:
			page_down();
			return;

		case CONTROL_CHAR:
			parse_control();
			return;

		case 18:	/* refresh - FIXME */
		case 12:
			doupdate();
			redrawwin(top); redrawwin(status); redrawwin(data);
			refresh(); doupdate();
			wrefresh(top); wrefresh(status);
			wrefresh(data);
			return;
			
		/* type in the character */
		default:
			type_char(c); return;
	}	
}

static void type_char(int c)
{
	byte a, b;
	int curs_x, curs_y;
	
	if (cursor_type == CURSOR_HEX) {
		if (!isxdigit(c)) return;
		if (isdigit(c))
			a = (byte) c - '0';
		else
			a = (byte) toupper(c) - 'A' + 10;

		curs_y = (cursor - page_start) / 16;
		curs_x = (cursor - page_start) % 16;
		curs_x = curs_x * 3 + 12;
		if (curs_x > 32) curs_x++;
		wmove(data, curs_y, curs_x);
		waddch(data, toupper(c) | A_BOLD);

		c = wgetch(data);
		while (c == ERR) {
			usleep(10000);
			c = wgetch(data);
		}
		if (!isxdigit(c)) {
			show_line(page_start, (cursor - page_start) / 16);
			move_cursor(cursor);
			return;
		}

		if (isdigit(c))
			b = (byte) c - '0';
		else
			b = (byte) toupper(c) - 'A' + 10;
		c = a * 16 + b;
	} 

	modified = 1;
	show_top(modified);
	text[cursor] = c; show_line(page_start, (cursor - page_start) / 16);
	move_cursor(cursor + 1);
}

void complain(int fatal, char *fmt, ...) {
	char tmp[1024];
	va_list ap;
	
	va_start(ap, fmt);
	vsnprintf(tmp, 1023, fmt, ap);
	tmp[1023] = 0;	
	if (fatal && screen) {
		wmove(status, 0, 0); wrefresh(status);
		endwin(); screen = NULL;
	}
	if (!screen) {
		fprintf(stderr, "%s\n", tmp);
		if (fatal) exit(EXIT_FAILURE);
		return;
	}
	print_status(tmp);
}

void print_status(char *txt)
{
	if (!screen) return;
	werase(status);
	wmove(status, 0, 0); waddstr(status, txt); wrefresh(status);
}

static void sigint_break(int x)
{
	signal(SIGINT, SIG_IGN);
	siglongjmp(func_env, 1);
}

int input_status(char *txt, char *buf, int len)
{
	int i;
	
	if (!screen) return -1;
	print_status(txt); nocbreak(); echo();

	signal(SIGINT, sigint_break);
	i = sigsetjmp(func_env, 1);
	if (!i) wgetnstr(status, buf, len); buf[len] = 0;
	signal(SIGINT, on_sigint);
	help_page = 0; print_help(1); cbreak(); noecho();
	if (!i)
		return 0;
	else
		return -1;
}

int ask_status(char *txt)
{
	int i = 0, c;

	if (!screen) return 0;
	print_status(txt); nodelay(status, TRUE);
	signal(SIGINT, sigint_break);
	i = sigsetjmp(func_env, 1);
	while (!i) {
		usleep(10000);
		c = wgetch(status);
		if (c == ERR) continue;
		c = toupper(c);
		if ((c == 'Y') || (c == 'N')) break;
	}
	signal(SIGINT, on_sigint);
	help_page = 0; print_help(1); nodelay(status, FALSE);
	if (i || (c == 'N')) return 0;
	return 1;
}

void clean_exit(void)
{
	wmove(status, 0, 0); wrefresh(status); endwin();
	printf("\nHave a nice Unix...\n");
	exit(EXIT_SUCCESS);
}

static void print_helpitem(byte *acc, byte *desc)
{
	waddstr(status, " [");
	while (*acc) {
		waddch(status, *acc | A_BOLD);
		acc++;
	}
	waddstr(status, " - "); waddstr(status, desc); waddstr(status, "]");
}

void print_help(int x)
{
	unsigned char *addr = "senko@fly.srk.fer.hr";
	int i;

	if (x) help_page = 0;
	
	switch(help_page) {
		case 0:
			werase(status); wmove(status, 0, 0);
			waddstr(status, "Ptolomei's Hex editor  --  Press ");
			waddch(status, '^' | A_BOLD);
			waddch(status, 'K' | A_BOLD);
			waddch(status, 'H' | A_BOLD);
			waddstr(status, " for help");
			wrefresh(status);
			move_cursor(cursor);
			break;

		case 1:
			werase(status); wmove(status, 0, 0);
			print_helpitem("^KE", "Edit");
			print_helpitem("^KD", "Save");
			print_helpitem("^KX", "Save & Exit");
			print_helpitem("^Z", "Go to Shell");
			print_helpitem("^C", "Quit");
			wrefresh(status); move_cursor(cursor); break;

		case 2:
			werase(status); wmove(status, 0, 0);
			print_helpitem("TAB", "Switch mode");
			print_helpitem("^U", "Page Up");
			print_helpitem("^V", "Page Down");
			print_helpitem("^KU", "Top");
			print_helpitem("^KV", "End");
			wrefresh(status); move_cursor(cursor); break;

		case 3:
			werase(status); wmove(status, 0, 0);
			print_helpitem("^KF", "Find");
			print_helpitem("^KN", "Find Next");
			print_helpitem("^KL", "Go to Offset");
			print_helpitem("^R", "Refresh Screen");
			wrefresh(status); move_cursor(cursor); break;

		case 4:
			werase(status); wmove(status, 0, 0);
			print_helpitem("^KB", "Begin Block");
			print_helpitem("^KK", "End Block");
			print_helpitem("^KC", "Copy");
			print_helpitem("^KR", "Insert");
			print_helpitem("^KW", "Write");
			wrefresh(status); move_cursor(cursor); break;

		case 5:
			werase(status); wmove(status, 0, 0);
			waddstr(status, "Copyright (C) 1997-2000  Senko Rasic <");
			for (i = 0; addr[i]; i++) waddch(status, addr[i] | A_BOLD);
			waddch(status, '>');
			wrefresh(status); move_cursor(cursor); break;
	}
	help_page++;
	
	if (help_page == 6) help_page = 0;
}

void update_screen(int x)
{
	if (x) {
		page_start = 0;
		sel_start = 1;
		sel_end = 0;
	}
	show_data(page_start);
	move_cursor(0);
}
